@@ -13,6 +13,12 @@ module Network.HTTP.Client.Manager
     , failedConnectionException
     , defaultManagerSettings
     , rawConnectionModifySocket
+    , proxyFromRequest
+    , noProxy
+    , useProxy
+    , proxyEnvironment
+    , proxyEnvironmentNamed
+    , defaultProxy
     ) where
 
 #if !MIN_VERSION_base(4,6,0)
@@ -30,9 +36,10 @@ import qualified Blaze.ByteString.Builder as Blaze
 
 import Data.Text (Text)
 import qualified Data.Text as T
+import Data.Text.Read (decimal)
 
 import Control.Monad.IO.Class (MonadIO, liftIO)
-import Control.Monad (unless, join)
+import Control.Monad (unless, join, when, void, mplus)
 import Control.Exception (mask_, SomeException, bracket, catch, throwIO, fromException, mask, IOException, Exception (..), handle)
 import Control.Concurrent (forkIO, threadDelay)
 import Data.Time (UTCTime (..), Day (..), DiffTime, getCurrentTime, addUTCTime)
@@ -47,6 +54,10 @@ import Network.HTTP.Types (status200)
 import Network.HTTP.Client.Types
 import Network.HTTP.Client.Connection
 import Network.HTTP.Client.Headers (parseStatusHeaders)
+import Control.Concurrent.MVar (MVar, takeMVar, tryPutMVar, newEmptyMVar)
+import System.Environment (getEnvironment)
+import qualified Network.URI as U
+import Control.Monad (guard)
 
 -- | A value for the @managerRawConnection@ setting, but also allows you to
 -- modify the underlying @Socket@ to set additional settings. For a motivating
@@ -57,8 +68,19 @@ rawConnectionModifySocket :: (NS.Socket -> IO ())
                           -> IO (Maybe NS.HostAddress -> String -> Int -> IO Connection)
 rawConnectionModifySocket = return . openSocketConnection
 
+-- | Same as @rawConnectionModifySocket@, but also takes in a chunk size.
+--
+-- Since 0.4.5
+rawConnectionModifySocketSize :: (NS.Socket -> IO ())
+                              -> IO (Int -> Maybe NS.HostAddress -> String -> Int -> IO Connection)
+rawConnectionModifySocketSize = return . openSocketConnectionSize
+
 -- | Default value for @ManagerSettings@.
 --
+-- Note that this value does /not/ have support for SSL/TLS. If you need to
+-- make any https connections, please use the http-client-tls package, which
+-- provides a @tlsManagerSettings@ value.
+--
 -- Since 0.1.0
 defaultManagerSettings :: ManagerSettings
 defaultManagerSettings = ManagerSettings
@@ -87,6 +109,9 @@ defaultManagerSettings = ManagerSettings
                     Nothing -> se
          in handle $ throwIO . wrapper
     , managerIdleConnectionCount = 512
+    , managerModifyRequest = return
+    , managerProxyInsecure = defaultProxy
+    , managerProxySecure = defaultProxy
     }
 
 takeSocket :: Manager -> ConnKey -> IO (Maybe Connection)
@@ -108,6 +133,7 @@ putSocket :: Manager -> ConnKey -> Connection -> IO ()
 putSocket man key ci = do
     now <- getCurrentTime
     join $ I.atomicModifyIORef (mConns man) (go now)
+    void $ tryPutMVar (mConnsBaton man) ()
   where
     go _ ManagerClosed = (ManagerClosed , connectionClose ci)
     go now mc@(ManagerOpen idleCount m)
@@ -145,14 +171,21 @@ addToList now maxCount x l@(Cons _ currCount _ _)
 -- Since 0.1.0
 newManager :: ManagerSettings -> IO Manager
 newManager ms = do
+    NS.withSocketsDo $ return ()
     rawConnection <- managerRawConnection ms
     tlsConnection <- managerTlsConnection ms
     tlsProxyConnection <- managerTlsProxyConnection ms
     mapRef <- I.newIORef $! ManagerOpen 0 Map.empty
+    baton <- newEmptyMVar
     wmapRef <- I.mkWeakIORef mapRef $ closeManager' mapRef
-    _ <- forkIO $ reap wmapRef
+
+    httpProxy <- runProxyOverride (managerProxyInsecure ms) False
+    httpsProxy <- runProxyOverride (managerProxySecure ms) True
+
+    _ <- forkIO $ reap baton wmapRef
     let manager = Manager
             { mConns = mapRef
+            , mConnsBaton = baton
             , mMaxConns = managerConnCount ms
             , mResponseTimeout = managerResponseTimeout ms
             , mRawConnection = rawConnection
@@ -161,12 +194,17 @@ newManager ms = do
             , mRetryableException = managerRetryableException ms
             , mWrapIOException = managerWrapIOException ms
             , mIdleConnectionCount = managerIdleConnectionCount ms
+            , mModifyRequest = managerModifyRequest ms
+            , mSetProxy = \req ->
+                if secure req
+                    then httpsProxy req
+                    else httpProxy req
             }
     return manager
 
 -- | Collect and destroy any stale connections.
-reap :: Weak (I.IORef ConnsMap) -> IO ()
-reap wmapRef =
+reap :: MVar () -> Weak (I.IORef ConnsMap) -> IO ()
+reap baton wmapRef =
     mask_ loop
   where
     loop = do
@@ -179,8 +217,13 @@ reap wmapRef =
     goMapRef mapRef = do
         now <- getCurrentTime
         let isNotStale time = 30 `addUTCTime` time >= now
-        toDestroy <- I.atomicModifyIORef mapRef (findStaleWrap isNotStale)
+        (newMap, toDestroy) <- I.atomicModifyIORef mapRef $ \m ->
+            let (newMap, toDestroy) = findStaleWrap isNotStale m
+             in (newMap, (newMap, toDestroy))
         mapM_ safeConnClose toDestroy
+        case newMap of
+            ManagerOpen _ m | not $ Map.null m -> return ()
+            _ -> takeMVar baton
         loop
     findStaleWrap _ ManagerClosed = (ManagerClosed, [])
     findStaleWrap isNotStale (ManagerOpen idleCount m) =
@@ -349,7 +392,7 @@ getConn req m
     -- https://github.com/snoyberg/http-client/issues/40#issuecomment-39117909
     | S8.null h = throwIO $ InvalidDestinationHost h
     | otherwise =
-        getManagedConn m (ConnKey connKeyHost connport (secure req)) $
+        getManagedConn m (ConnKey connKeyHost connport (host req) (port req) (secure req)) $
             wrapConnectExc $ go connaddr connhost connport
   where
     h = host req
@@ -377,7 +420,100 @@ getConn req m
                         , " HTTP/1.1\r\n\r\n"
                         ]
                     parse conn = do
-                        sh@(StatusHeaders status _ _) <- parseStatusHeaders conn
+                        sh@(StatusHeaders status _ _) <- parseStatusHeaders conn Nothing Nothing
                         unless (status == status200) $
                             throwIO $ ProxyConnectException ultHost ultPort $ Left $ S8.pack $ show sh
                  in mTlsProxyConnection m connstr parse (S8.unpack ultHost)
+
+-- | Get the proxy settings from the @Request@ itself.
+--
+-- Since 0.4.7
+proxyFromRequest :: ProxyOverride
+proxyFromRequest = ProxyOverride $ const $ return id
+
+-- | Never connect using a proxy, regardless of the proxy value in the @Request@.
+--
+-- Since 0.4.7
+noProxy :: ProxyOverride
+noProxy = ProxyOverride $ const $ return $ \req -> req { proxy = Nothing }
+
+-- | Use the given proxy settings, regardless of the proxy value in the @Request@.
+--
+-- Since 0.4.7
+useProxy :: Proxy -> ProxyOverride
+useProxy p = ProxyOverride $ const $ return $ \req -> req { proxy = Just p }
+
+-- | Get the proxy settings from the default environment variable (@http_proxy@
+-- for insecure, @https_proxy@ for secure). If no variable is set, then fall
+-- back to the given value. @Nothing@ is equivalent to 'noProxy', @Just@ is
+-- equivalent to 'useProxy'.
+--
+-- Since 0.4.7
+proxyEnvironment :: Maybe Proxy -- ^ fallback if no environment set
+                 -> ProxyOverride
+proxyEnvironment mp = ProxyOverride $ \secure ->
+    envHelper (envName secure) $ maybe EHNoProxy EHUseProxy mp
+
+envName :: Bool -- ^ secure?
+        -> Text
+envName False = "http_proxy"
+envName True = "https_proxy"
+
+-- | Same as 'proxyEnvironment', but instead of default environment variable
+-- names, allows you to set your own name.
+--
+-- Since 0.4.7
+proxyEnvironmentNamed
+    :: Text -- ^ environment variable name
+    -> Maybe Proxy -- ^ fallback if no environment set
+    -> ProxyOverride
+proxyEnvironmentNamed name =
+    ProxyOverride . const . envHelper name
+                  . maybe EHNoProxy EHUseProxy
+
+-- | The default proxy settings for a manager. In particular: if the @http_proxy@ (or @https_proxy@) environment variable is set, use it. Otherwise, use the values in the @Request@.
+--
+-- Since 0.4.7
+defaultProxy :: ProxyOverride
+defaultProxy = ProxyOverride $ \secure ->
+    envHelper (envName secure) EHFromRequest
+
+data EnvHelper = EHFromRequest
+               | EHNoProxy
+               | EHUseProxy Proxy
+
+envHelper :: Text -> EnvHelper -> IO (Request -> Request)
+envHelper name eh = do
+    env <- getEnvironment
+    case lookup (T.unpack name) env of
+        Nothing -> return $
+            case eh of
+                EHFromRequest -> id
+                EHNoProxy     -> \req -> req { proxy = Nothing }
+                EHUseProxy p  -> \req -> req { proxy = Just p  }
+        Just str -> do
+            let invalid = throwIO $ InvalidProxyEnvironmentVariable name (T.pack str)
+            p <- maybe invalid return $ do
+                uri <- case U.parseURI str of
+                    Just u | U.uriScheme u == "http:" -> return u
+                    _ -> U.parseURI $ "http://" ++ str
+
+                guard $ U.uriScheme uri == "http:"
+                guard $ null (U.uriPath uri) || U.uriPath uri == "/"
+                guard $ null $ U.uriQuery uri
+                guard $ null $ U.uriFragment uri
+
+                auth <- U.uriAuthority uri
+                guard $ null $ U.uriUserInfo auth
+
+                port <-
+                    case U.uriPort auth of
+                        "" -> Just 80
+                        ':':rest ->
+                            case decimal $ T.pack rest of
+                                Right (p, "") -> Just p
+                                _ -> Nothing
+                        _ -> Nothing
+
+                Just $ Proxy (S8.pack $ U.uriRegName auth) port
+            return $ \req -> req { proxy = Just p }
